""" Usage examples of the MyRIO class: MXP board tests

    This examples shows how to use the electronic components
    we already have on our MXP boards. It serves as example of
    more sophisticated functions we added to myRIO_base.

    Our MXP cards have an RGB LED connected to DIO_2:0 and 
    two push buttons on DIO_4:3. We also have an NTC temperature
    sensor in AI0 and an LDR light sensor in AI1. We use these
    channels in our examples. The port that we use most is the 
    A port, so we set it as default on our package. 

    Last update: 2024/03/14 Aitzol Ezeiza Ramos (UPV/EHU)
"""

import myRIO_base as myRIO
from time import sleep

myrio1 = myRIO.MyRIO()

print("Read temperature from MXP port A, channel 0:")
print(myrio1.read_MXP_temperature())
print("Push the first button to continue (DIO3)")
while not (myrio1.read_MXP_button(1)):
    sleep(0.1)
print("Read luminosity from MXP port A, channel 1:")
print(myrio1.read_MXP_luminosity())
print("Push the second button to continue (DIO4)")
while not (myrio1.read_MXP_button(2)):
    sleep(0.1)
print("RGB LED in MXP port A: RED")
myrio1.write_MXP_RGB_LED(myRIO.RED)
sleep(1)
print("RGB LED in MXP port A: GREEN")
myrio1.write_MXP_RGB_LED(myRIO.GREEN)
sleep(1)
print("RGB LED in MXP port A: BLUE")
myrio1.write_MXP_RGB_LED(myRIO.BLUE)
sleep(1)
print("RGB LED in MXP port A: OFF")
myrio1.write_MXP_RGB_LED(myRIO.RGB_OFF)


""" Usage examples of the MyRIO class: digital inputs and outputs

    This examples shows how to use the DIO ports and the
    onboard button and LEDs.

    We have a set of myRIOs in our facilities that we complement
    with some MXP cards. These cards have an RGB LED connected to
    DIO_2:0 and two push buttons on DIO_4:3. We use these channels
    in our examples. The port that we use most is the A port, so
    we set it as default on our package. 

    Last update: 2024/03/07 Aitzol Ezeiza Ramos (UPV/EHU)
"""

import myRIO_base
import time

myrio1 = myRIO_base.MyRIO()

# Read digital port (the whole port), one channel, and the onboard button

print('Digital port A:')
print(myrio1.read_digital_port(port='A'))
print('Digital channel 3, port A:')
print(myrio1.read_digital_input(3, port='A'))
print('onboard button:')
print(myrio1.read_button())

# Write onboard LEDs, using an integer and using a list of Booleans

print('Switching on the onboard LEDS (15 in binary)')
myrio1.write_leds_integer(7)
time.sleep(2)
print('Switching off the onboard LEDS (0 in binary)')
myrio1.write_leds_integer(0)
time.sleep(2)
print('Switching on the onboard LEDS 1 and 4 ([True, False, False, True])')
myrio1.write_leds_booleans([True, False, False, True])
time.sleep(2)
print('Switching off all the onboard LEDS ([False, False, False, False])')
myrio1.write_leds_booleans([False, False, False, False])
time.sleep(2)

# myRIO digital output test (the whole port at once, and then one by one)

print('Switching on the three RGB LEDS (7 in binary)')
myrio1.write_digital_port(7)
time.sleep(2)
print('Switching off the three RGB LEDS and all the other outputs (0,0 in binary)')
myrio1.write_digital_port(0,0)
time.sleep(2)

print('Switching on the green LED (channel 0)')
myrio1.write_digital_output(0,True)
time.sleep(1)
print('Switching on the red LED (channel 1)')
myrio1.write_digital_output(1,True)
time.sleep(1)
print('Switching on the blue LED (channel 2)')
myrio1.write_digital_output(2,True)
time.sleep(1)

print('Switching off the three RGB LEDS one by one')
myrio1.write_digital_output(0,False)
time.sleep(1)
myrio1.write_digital_output(1,False)
time.sleep(1)
myrio1.write_digital_output(2,False)
time.sleep(1)

print('One color at a time (3 reps)')
for i in range(3):
    myrio1.write_digital_port(myRIO_base.RED)
    time.sleep(1)
    myrio1.write_digital_port(myRIO_base.GREEN)
    time.sleep(1)
    myrio1.write_digital_port(myRIO_base.BLUE)
    time.sleep(1)

print('End of digital tests')
myrio1.write_digital_port(myRIO_base.RGB_OFF)


import myRIO_API_client as myAPIc
from time import sleep

myRIO = myAPIc.MyRIO_API_Client()
# myRIO = MyRIO_API_Client(ip_address='172.22.11.3',port:8080)

print('Accelerometer:')
print(myRIO.get_onboard_accelerometer())
print('Onboard button:')
print(myRIO.get_onboard_button())
print('Digital Input 3:')
print(myRIO.get_digital_input(3))
print('Digital Input 0 (port B):')
print(myRIO.get_digital_input(3,port_in='B'))
print('Analog Input 1:')
print(myRIO.get_analog_input(1))
print('Onboard LEDS (all ON)')
myRIO.set_onboard_leds(15)
sleep(1)
print('Onboard LEDS (all OFF)')
myRIO.set_onboard_leds(0)
sleep(1)
print('Digital Output 0 ON')
myRIO.set_digital_output(0,1)
sleep(1)
print('Digital Output 0 OFF')
myRIO.set_digital_output(0,0)
sleep(1)
print('Analog Output 0 5.0V')
myRIO.set_analog_output(0,5.0)
print('Analog Input 2: (the same if wired)')
print(myRIO.get_analog_input(2))
sleep(1)
print('Analog Output 0 0.0V')
myRIO.set_analog_output(0,0.0)
print('Analog Input 2: (the same if wired)')
print(myRIO.get_analog_input(2))
print('MXP temperature')
print(myRIO.get_mxp_temperature())
print('MXP luminosity')
print(myRIO.get_mxp_luminosity())
print('MXP buttons')
print(myRIO.get_mxp_button(1))
print(myRIO.get_mxp_button(2))
print('MXP RGB LED set GREEN')
myRIO.set_mxp_rgb_color(1)
sleep(1)
print('MXP RGB LED set RED')
myRIO.set_mxp_rgb_color(2)
sleep(1)
print('MXP RGB LED set BLUE')
myRIO.set_mxp_rgb_color(4)
sleep(1)
print('MXP RGB LED set WHITE')
myRIO.set_mxp_rgb_color(7)
sleep(1)
print('MXP RGB LED set OFF')
myRIO.set_mxp_rgb_color(0)



""" Usage examples of the MyRIO class: analog inputs and outputs

    This examples shows how to use the analog inputs and outputs
    and the onboard Accelerometer.

    We have a set of myRIOs in our facilities that we complement
    with some MXP cards. These cards have a temperature sensor in
    AI0 and a Light sensor in AI1. We use these channels
    in our examples. The port that we use most is the A port, so
    we set it as default on our package. 

    Last update: 2024/03/07 Aitzol Ezeiza Ramos (UPV/EHU)
"""

import myRIO_base
import time

myrio1 = myRIO_base.MyRIO()

# Read onboard accelerometer values
print('Accelerometer test (5 seconds)')
for i in range(10):
    print(myrio1.read_analog_accelerometer())
    time.sleep(0.5)
print('End of Accelerometer test')
time.sleep(2)

# Read analog inputs
print('Analog Inputs test (5 seconds) (values in volts)')
for i in range(10):
    print(str(myrio1.read_analog_input(0))+" "+str(myrio1.read_analog_input(1)))
    time.sleep(0.5)
print('End of Analog Inputs test')
time.sleep(2)

# Test the analog output wiring it to AI2
print('myRIO analog output test - wire AO0 to AI2 and press the onboard button')

while not myrio1.read_button():
    time.sleep(0.1)

print('AO0 write: 5.0')
myrio1.write_analog_output(channel=0,value=5.0)
time.sleep(1)
print('AI2 read: ', str(myrio1.read_analog_input(channel=2)))
print('AO0 write: 2.5')
myrio1.write_analog_output(channel=0,value=2.5)
time.sleep(1)
print('AI2 read: ', str(myrio1.read_analog_input(channel=2)))
print('AO0 write: 0.0')
myrio1.write_analog_output(channel=0,value=0.0)
time.sleep(1)
print('AI2 read: ', str(myrio1.read_analog_input(channel=2)))
print('End of Analog Output test')


""" myRIO API: A RESTful API server for the myRIO

    Last update: 2024/03/14 Aitzol Ezeiza Ramos UPV/EHU

    This library uses Flask and Waitress to create an API
    based on myRIO_library. Not all the functions are
    served this way, but the basics are there:
        · Digital Inputs and Outputs
        · Analog Inputs and Outputs
        · onboard button and LEDs
        · onboard accelerometer
        - MXP board components
    The default port is 8080. Some examples of API calls
    (using curl) would be the following:

    curl -X POST http://172.22.11.2:8080/digital_output/2/1

    This turns on the digital output DIO2 on the default MXP port (A)
    If you use Windows Powershell, you should replace the -X with -method X

    curl -method POST http://172.22.11.2:8080/digital_output/2/1

    You can change the default port with a parameter:

    curl -X GET http://172.22.11.2:8080/digital_input/3?port=B

    Read the examples un the examples folder for more info.
    """


from flask import Flask, jsonify, request
import myRIO_base

DEFAULT_HTTP_PORT = 8080

app = Flask(__name__)

@app.before_first_request
def setup():
    """ We need to create a session before any API call
        We use a global handler for that.
    """
    global myrio_handler
    myrio_handler = myRIO_base.MyRIO()

@app.route('/digital_input/<int:channel_in>', methods=['GET'])
def get_digital_input(channel_in: int):
    """ Returns the value (true/false) of a digital input """
    port_in = request.args.get('port', default='A', type=str)
    value = myrio_handler.read_digital_input(channel=channel_in,port=port_in)
    return jsonify({'value': value})

@app.route('/digital_output/<int:channel_in>/<int:value_in>', methods=['POST'])
def set_digital_output(channel_in: int, value_in: int):
    """ Sets the value (true 1, false 0) of a digital output """
    port_in = request.args.get('port', default='A', type=str)
    myrio_handler.write_digital_output(channel=channel_in,
                                       port=port_in,
                                       value=bool(value_in))
    return jsonify({'success': True})

@app.route('/analog_input/<int:channel_in>', methods=['GET'])
def get_analog_input(channel_in: int):
    """ Returns the value (volts in float type) of an analog input """
    port_in = request.args.get('port', default='A', type=str)
    value = myrio_handler.read_analog_input(channel=channel_in,port=port_in)
    return jsonify({'value': value})

@app.route('/analog_output/<int:channel_in>/<float:value_in>', methods=['POST'])
def set_analog_output(channel_in: int, value_in: float):
    """ Sets the value (volts in float type) of an analog output """
    port_in = request.args.get('port', default='A', type=str)
    myrio_handler.write_analog_output(channel=channel_in, port=port_in, value=value_in)
    return jsonify({'success': True})

# Define routes for the onboard inputs and outputs
@app.route('/onboard_button', methods=['GET'])
def get_onboard_button():
    """ Returns the value (true/false) of the onboard button """
    value = myrio_handler.read_button()
    return jsonify({'value': value})

@app.route('/onboard_leds/<int:value_in>', methods=['POST'])
def set_onboard_leds(value_in: int):
    """ Sets the value (0..15 integer) of the onboard LEDs """
    myrio_handler.write_leds_integer(value_in)
    return jsonify({'success': True})

@app.route('/onboard_accelerometer', methods=['GET'])
def get_onboard_accelerometer():
    """ Returns the value (x, y, z floats) of the onboard accelerometer """
    values = myrio_handler.read_analog_accelerometer()
    response_data = {
        'x': values[0],
        'y': values[1],
        'z': values[2]
    }
    return jsonify(response_data)

# MXP board components: buttons, RGB LED, Temperature sensor, Light sensor

@app.route('/mxp_button/<int:button_in>', methods=['GET'])
def get_mxp_button(button_in: int):
    """ Returns the value (true/false) of one of the buttons.
        We expect 1 or 2, first (black) and second (white).
    """
    port_in = request.args.get('port', default='A', type=str)
    value = myrio_handler.read_MXP_button(button=button_in,port=port_in)
    return jsonify({'value': value})

@app.route('/mxp_rgb_color/<int:color_in>', methods=['POST'])
def set_mxp_rgb_color(color_in: int):
    """ Sets the color (0 to 7) of the MXP RGB LED
        Remember that the order is G(0)R(1)B(2)
        Green is 1, Red is 2, Blue is 4
        White is 7 and Off is 0.
    """
    port_in = request.args.get('port', default='A', type=str)
    myrio_handler.write_MXP_RGB_LED(color=color_in, port=port_in)
    return jsonify({'success': True})

@app.route('/mxp_temperature', methods=['GET'])
def get_mxp_temperature():
    """ Returns the temperature (degrees in float type) of
        the NTC temperature sensor (MXP board)
    """
    port_in = request.args.get('port', default='A', type=str)
    value = myrio_handler.read_MXP_temperature(port=port_in)
    return jsonify({'value': value})

@app.route('/mxp_luminosity', methods=['GET'])
def get_mxp_luminosity():
    """ Returns the luminosity (percentage in float type) of
        the LDR light sensor (MXP board)
    """
    port_in = request.args.get('port', default='A', type=str)
    value = myrio_handler.read_MXP_luminosity(port=port_in)
    return jsonify({'value': value})


if __name__ == '__main__':
    from waitress import serve
    serve(app, host='0.0.0.0', port=DEFAULT_HTTP_PORT)
    # 0.0.0.0 means it will be served in all network interfaces

    # Instead of serving, you can debug the application
    # using Flask's server
    # app.run(debug=True)



""" myRIO_library - a library for working with NI myRIO in Python

This library is an improvement over nifpga, a Python library that
gives access to the FPGA registers of NI targets with FPGA.

In this library, we have created some support functions and a class
named MyRIO.
"""

from nifpga import Session
from typing import Tuple
import ctypes
import pkg_resources


# RGB constants

RED = 2
GREEN = 1
BLUE = 4
WHITE = 7
RGB_OFF = 0


# Support functions. They are not part of the class MyRIO, but they are used by it.
# Feel free to use them in your own programs.

def u8_to_bits (u8_number: int) -> [bool]:
    """ This function converts u8 values to an array of Booleans """
    mask = 1
    one_by_one = []
    for i in range(8):
        one_by_one.append(bool((u8_number & mask) != 0))
        mask = mask * 2
    return(one_by_one)

def only_one_bit_on (bit_number: int, input_number: int=0) -> int:
    """ This function switches on only one bit on a u8 integer """
    u8_number = 0
    if bit_number <= 7:
        u8_number = (1 << bit_number)
    return input_number | u8_number
    #TODO raise a warning if bit_number is too big

def only_one_bit_off (bit_number: int, input_number: int=0) -> int:
    """ This function switches on only one bit off a u8 integer """
    
    if bit_number <= 7:
        mask = ~(1 << bit_number)
        return input_number & mask
    else:
        return input_number        #TODO raise a warning if bit_number is too big

def raw_to_volts_AB (raw: int) -> float:
    """ This function converts the raw value 
    from myRIO MXP (ports A and B) Analog Input channels to volts
    """

    return raw * 0.001220703

def raw_to_volts_C (raw: int) -> float:
    """ This function converts the raw value 
    from myRIO MSP (port C) Analog Input channels to volts
    """
    
    return raw * 0.004882813

def raw_to_volts_audio (raw: int) -> float:
    """ This function converts the raw value 
    from myRIO Audio Input channels to volts
    """
    
    return raw * 0.001220703

def volts_to_raw_AB (volts: float) -> int:
    """ This function converts values in Volts 
    to raw values for myRIO MXP (ports A and B) Analog Output channels
    """
    
    return int(volts / 0.001220703)

def volts_to_raw_C (volts: float) -> int:
    """ This function converts values in Volts 
    to raw values for myRIO MSP (port C) Analog Output channels
    """ 

    signed_value = int(volts / 0.004882813)
    unsigned_value = ctypes.c_uint16(signed_value).value
    return unsigned_value

def volts_to_raw_audio (volts: float) -> int:
    """ This function converts values in Volts 
    to raw values for Audio Output channels
    """

    signed_value = int(volts / 0.001220703)
    unsigned_value = ctypes.c_uint16(signed_value).value
    return unsigned_value

def volts_to_temperature (volts: float) -> float:
    """ This function converts values in Volts 
    to temperature in Celsius degrees
    """
    return ((volts-1.7)*10/-0.3)+20

def volts_to_luminosity (volts: float) -> float:
    """ This function converts values in Volts 
    to luminosity in percentage (0-100%)
    """
    return 100-(volts*30.3)

def extract_waveform_from_csv_file(file_name: str) -> [int]:
    """ This function extracts a stereo waveform from a .csv file
    and returns it as a list of I16 integers.
    The .csv format is very basic for an audio file, but
    the myRIO has memory limitations, so the audio files
    must be mono and quite shorts.
    We have tried the wave library from Python, but it does not work
    properly with the myRIO, so we have used the csv format that 
    can be easily generated (we used a simple LabVIEW program for that).
    """
    # Open and read a csv file 
    with open(file_name, 'r') as file:
        data = file.readlines()     # Read all the lines of the file
        waveform = []               # Create an empty list for the waveform
        for line in data:           # Read line by line
            waveform.append(volts_to_raw_audio(float(line))) # Append the integer value to the list

    return waveform                # Return the list



class MyRIO:
    """ Class MyRIO: class for accessing NI myRIO inputs and outputs using nifgpa

    NI myRIO is a programmable device with digital and analog inputs and outputs.
    It is usually programmed in LabVIEW, but it can be programmed in Python too.
    """

    __session = None

    def __init__(self, session_bitfile="", session_resource = "RIO0"):
        """ when an instance is created, an nifpga session is created. """
        if session_bitfile == "":
            session_bitfile = pkg_resources.resource_filename('myRIO_base', 'data/Default.lvbitx')

        self.__session = Session(session_bitfile, session_resource)
        sys_handler = self.__session.registers['SYS.RDY']
        while not sys_handler.read():       #TODO: What if the system is never ready?
            time.sleep(0.1)

    def check_if_ready(self) -> bool:
        """ checks if system is ready """
        sys_handler = self.__session.registers['SYS.RDY']
        return sys_handler.read()

    def __del__(self):
        """ This function closes the session with the myRIO FPGA """
        self.__session.close()

    def set_DIO_mask(self, mask_low: int =7, mask_high: int =0, port: str ='A'):
        """ sets the DIO mask for defining the direction (IN/OUT) of the channels

        low 7, high 0 is the default for the design of our current MXP cards
        """

        dir_string_low = 'DIO.' + port + '_' + '7:0' + '.DIR'
        dir_string_high = 'DIO.' + port + '_' + '15:8' + '.DIR'
        mask_handler_low = self.__session.registers[dir_string_low]
        mask_handler_high = self.__session.registers[dir_string_high]
        mask_handler_low.write(mask_low)    # 7 is 00000111 where 1 is OUT
        mask_handler_high.write(mask_high)    # 0 is 00000000 where 0 is IN


    def read_analog_input(self, channel: int, port: str ='A') -> float:
        """ returns the value in volts of one of the AI channels (default port: A) """
    
        channel_string = 'AI.' + port + '_' + str(channel) + '.VAL'
        channel_handler = self.__session.registers[channel_string]
        if port=='A' or port == 'B':
            return raw_to_volts_AB(channel_handler.read())
        elif port=='C':
            return raw_to_volts_C(channel_handler.read())
        elif port=='AudioIn_L' or port =='AudioIn_R':
            return raw_to_volts_audio(channel_handler.read())
        else:
            #TODO how to report the exception (port name error)
            return None

    def read_MXP_temperature(self, channel: int=0, port: str ='A') -> float:
        """ returns the temperature in Celsius degrees of one of the 
            AI channels (default channel:0, default port: A)
        """
        channel_string = 'AI.' + port + '_' + str(channel) + '.VAL'
        channel_handler = self.__session.registers[channel_string]
        if port=='A' or port == 'B':
            return volts_to_temperature(raw_to_volts_AB(channel_handler.read()))
        else:
            #TODO how to report the exception (port name error)
            return None

    def read_MXP_luminosity(self, channel: int =1, port: str ='A') -> float:
        """ returns the luminosity in percentage of one of the 
            AI channels (default channel:1, default port: A)
        """
        channel_string = 'AI.' + port + '_' + str(channel) + '.VAL'
        channel_handler = self.__session.registers[channel_string]
        if port=='A' or port == 'B':
            return volts_to_luminosity(raw_to_volts_AB(channel_handler.read()))
        else:
            #TODO how to report the exception (port name error)
            return None

    def read_digital_input(self, channel: int, port: str ='A') -> bool:
        """ returns the Boolean value of one of the DIO input channels (default port: A) """

        # First, check if the function is asking for a channel 
        # in the low byte (7:0) or in the high byte (15:8)
        if channel<8 :
            channel_string = 'DIO.' + port + '_7:0.IN'
            array_index = int(channel)
        else:
            channel_string = 'DIO.' + port + '_15:8.IN'
            # The length of the array is 8, so the index should be in the 7:0 range
            array_index = channel-8

        channel_handler = self.__session.registers[channel_string]
        raw_value = channel_handler.read()
        bool_array = u8_to_bits(raw_value)
        return bool_array[array_index]

    def read_digital_port(self, port: str ='A') -> [bool]:
        """ returns the Boolean values of the whole port (default port: A) """

        channel_string_low = 'DIO.' + port + '_7:0.IN'
        channel_string_high = 'DIO.' + port + '_15:8.IN'
        channel_handler_low = self.__session.registers[channel_string_low]
        channel_handler_high = self.__session.registers[channel_string_high]
        raw_value_low = channel_handler_low.read()
        raw_value_high = channel_handler_high.read()
        bool_array = u8_to_bits(raw_value_low) + u8_to_bits(raw_value_high)
        return bool_array

    def read_MXP_button(self, button: int =1, port: str ='A') -> bool:
        """ returns the Boolean value of one of the MXP buttons (default port: A) 
            We spect 1 for the first button, 2 for the second one.
            Most of our cards have a black button first, a white button second.
        """
        channel_string = 'DIO.' + port + '_7:0.IN'
        array_index = int(button+2)
        channel_handler = self.__session.registers[channel_string]
        raw_value = channel_handler.read()
        bool_array = u8_to_bits(raw_value)
        return bool_array[array_index]

    def read_button(self) -> bool:
        """ returns the Boolean value of the myRIO onboard button """

        channel_string = 'DI.BTN'
        channel_handler = self.__session.registers[channel_string]
        raw_value = channel_handler.read()
        return bool(raw_value)

    def read_analog_accelerometer(self) -> Tuple[float, float, float]:
        """ returns the x, y, and z values in Gs of the onboard Accelerometer """
    
        channel_string_x = 'ACC.X.VAL'
        channel_string_y = 'ACC.Y.VAL'
        channel_string_z = 'ACC.Z.VAL'
        channel_handler_x = self.__session.registers[channel_string_x]
        channel_handler_y = self.__session.registers[channel_string_y]
        channel_handler_z = self.__session.registers[channel_string_z]
        raw_value_x = channel_handler_x.read()
        raw_value_y = channel_handler_y.read()
        raw_value_z = channel_handler_z.read()

        if raw_value_x & (1 << 15):
        # Perform two's complement conversion for negative numbers
            raw_value_x = raw_value_x - (1 << 16)
        x_value = float(raw_value_x) / 256.0

        if raw_value_y & (1 << 15):
        # Perform two's complement conversion for negative numbers
            raw_value_y = raw_value_y - (1 << 16)
        y_value = float(raw_value_y) / 256.0

        if raw_value_z & (1 << 15):
        # Perform two's complement conversion for negative numbers
            raw_value_z = raw_value_z - (1 << 16)
        z_value = float(raw_value_z) / 256.0

        return x_value, y_value, z_value

    def write_leds_integer(self, raw_value: int):
        """ changes the state of the myRIO onboard LEDs using an integer value """

        value = max(0, min(raw_value, 15))      #TODO: raising a warning when out of range?
        channel_string = 'DO.LED3:0'
        channel_handler = self.__session.registers[channel_string]
        channel_handler.write(value)

    def write_leds_booleans(self, boolean_values: [bool]):
        """ changes the state of the myRIO onboard LEDs using Booleans """

        raw_value = 0
        j = 1
        for i in range(4):
            raw_value = raw_value + boolean_values[i]*j
            j = j*2

        channel_string = 'DO.LED3:0'
        channel_handler = self.__session.registers[channel_string]
        channel_handler.write(raw_value)
        return None

    def write_digital_output(self,
                             channel: int,
                             value: bool,
                             port: str ='A',
                             mask_low: int =7,
                             mask_high: int =0):
        """ writes a Boolean value on one of the DIO output channels (default port: A)
        The defaults are set for the design of our current MXP cards: 
        RGB LEDs at channels 0,1,2 (G,R,B) and buttons at channels 3,4. Rest unused.
        """
        self.set_DIO_mask(port=port, mask_low=mask_low, mask_high=mask_high)

        # Check if the function is asking for a channel 
        # in the low byte (7:0) or in the high byte (15:8)
        if 0 <= channel <= 7 :
            channel_string = 'DIO.' + port + '_7:0.OUT'
            mask_string = 'DIO.' + port + '_7:0.DIR'
        elif 8 <= channel <= 15:
            channel_string = 'DIO.' + port + '_15:8.OUT'
            mask_string = 'DIO.' + port + '_15:8.DIR'
            channel = channel-8
        else:
            #TODO raise an error if channel is out of range
            return None

        mask_handler = self.__session.registers[mask_string]
        saved_mask = mask_handler.read()
        new_mask = only_one_bit_on(channel, saved_mask)
        mask_handler.write(new_mask)

        channel_handler = self.__session.registers[channel_string]
        saved_value = channel_handler.read()

        if value:
            value_to_be_written = only_one_bit_on(channel,saved_value)
        else:
            value_to_be_written = only_one_bit_off(channel,saved_value)

        channel_handler.write(value_to_be_written)

        mask_handler.write(saved_mask)   # restore original DIR mask

    def write_digital_port(self, 
                           value_low: int, 
                           value_high: int=255, 
                           port: str ='A',
                           mask_low: int =7,
                           mask_high: int =0):
        """ writes integer low and high values on DIO output channels (default port: A)
        The defaults are set for the design of our current MXP cards: 
        RGB LEDs at channels 0,1,2 (G,R,B) and buttons at channels 3,4. Rest unused.
        """
        
        self.set_DIO_mask(port=port, mask_low=mask_low, mask_high=mask_high)

        channel_string_low = 'DIO.' + port + '_7:0.OUT'
        channel_string_high = 'DIO.' + port + '_15:8.OUT'
        channel_handler_low = self.__session.registers[channel_string_low]
        channel_handler_high = self.__session.registers[channel_string_high]
        channel_handler_low.write(value_low)
        channel_handler_high.write(value_high)
 
    def write_MXP_RGB_LED(self, color: int, port: str ='A'):
        """ writes a color on the myRIO MXP RGB LED (default port: A) """

        self.set_DIO_mask(port=port)    # default mask is OK for the RGB LED
        channel_string_low = 'DIO.' + port + '_7:0.OUT'
        channel_handler_low = self.__session.registers[channel_string_low]
        channel_handler_low.write(color)
 
    def write_analog_output(self, channel: int, value: float, port: str ='A'):
        """ writes a value (in volts) on an AO channel (default port: A) """
    
        channel_string = 'AO.' + port + '_' + str(channel) + '.VAL'
        channel_handler = self.__session.registers[channel_string]
        if port=='A' or port == 'B':
            raw_value = volts_to_raw_AB(value)
        elif port=='C':
            raw_value = volts_to_raw_C(value)
        elif port=='AudioOut_L' or port =='AudioOut_R':
            raw_value = volts_to_raw_audio(value)
        else:
            #TODO how to report the exception (port name error)
            raw_value = 0
        channel_handler.write(raw_value)
        go_handler = self.__session.registers['AO.SYS.GO']
        go_handler.write(True)

    def play_waveform(self, waveform: [int]):
        """ plays a waveform on the myRIO Audio Output channels """
    
        channel_string_left = 'AO.AudioOut_L.VAL'
        channel_string_right = 'AO.AudioOut_R.VAL'
        channel_handler_left = self.__session.registers[channel_string_left]
        channel_handler_right = self.__session.registers[channel_string_right]

        go_handler = self.__session.registers['AO.SYS.GO']

        for i in range(len(waveform)):
            channel_handler_left.write(waveform[i])
            channel_handler_right.write(waveform[i])
            go_handler.write(True)

if __name__ == "__main__":
    print("This is a library for working with NI myRIO in Python")
    print("It is not intended to be run directly, but to be imported in other programs.")
    print("Please, see the documentation for more information.")
    from time import sleep
    myrio1 = MyRIO()

    
    print("Read digital port A:")
    print(myrio1.read_digital_port(port='A'))
    print("Read temperature from MXP port A, channel 0:")
    print(myrio1.read_MXP_temperature())
    print("Read luminosity from MXP port A, channel 1:")
    print(myrio1.read_MXP_luminosity())
    print("RGB LED in MXP port A: RED")
    myrio1.write_MXP_RGB_LED(RED)
    sleep(1)
    print("RGB LED in MXP port A: GREEN")
    myrio1.write_MXP_RGB_LED(GREEN)
    sleep(1)
    print("RGB LED in MXP port A: BLUE")
    myrio1.write_MXP_RGB_LED(BLUE)
    sleep(1)
    print("RGB LED in MXP port A: OFF")
    myrio1.write_MXP_RGB_LED(RGB_OFF)

    
    # Play a simple waveform (2024/04/10)
    print("Playing a simple waveform")
    csv_file = pkg_resources.resource_filename('myRIO_base', 'examples/PinkPanther.csv')
    my_waveform = extract_waveform_from_csv_file(csv_file)
    myrio1.play_waveform(my_waveform)



        
""" TODO

There are some extra features that we do not cover.
They are interesting, but are not so commonly used, and given
their complexity, we leave them for future development.
The features we did not cover are:
1.-PWM and ENC
2.-I2C and SPI
3.-IRQs

It would be interesting too to improve the audio functions.
"""

""" Credits
This library has been developed from scratch by Aitzol Ezeiza Ramos
from the University of the Basque Country (UPV/EHU)
It is strongly based on nifpga the basic library for accessing the
FPGA on NI RIO devices.

https://github.com/ni/nifpga-python/

We also use typing and ctypes.

First version: 2024/02/28
Current version: 2024/03/14
"""



""" Usage examples of the MyRIO class: play waveform

    This example shows how to play a waveform using the MyRIO class.

    Playing wav files is not supported in the myRIO library. However,
    there is a rudimentary way to play a waveform using raw data.

    Last update: 2024/04/10 Aitzol Ezeiza Ramos (UPV/EHU)
"""

import myRIO_base as myRIO
import pkg_resources

myrio1 = myRIO.MyRIO()

# Play a simple waveform using raw data
print("Playing a simple waveform")
csv_file = pkg_resources.resource_filename('myRIO_base', 'examples/PinkPanther.csv')
print("Waveform file: ", csv_file)
my_waveform = extract_waveform_from_csv_file(csv_file)
myrio1.play_waveform(my_waveform)


from setuptools import setup, find_packages
# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
name='myRIO_library',
version='1.2.1',
author='Aitzol Ezeiza Ramos',
author_email='aitzol.ezeiza@ehu.eus',
description='A library to control the myRIO board from National Instruments',
long_description=long_description,
long_description_content_type='text/markdown',

packages=find_packages('src'),
package_dir={'': 'src'},
include_package_data=True,
package_data={
        'myRIO_base': ['data/*'],
        'myRIO_base': ['examples/*'],
        'myRIO_API': ['examples/*'],
        'myRIO_API_client': ['examples/*'],
    },
classifiers=[
'Programming Language :: Python :: 3',
'License :: OSI Approved :: MIT License',
'Operating System :: Other OS',
'Operating System :: OS Independent',
],
python_requires='>=3.5',
)